import random


PLATFORMS: list[str] = [
    "X11; Linux x86_64",
    "Windows NT 6.1; WOW64",
    "Windows NT 10.0; Win64; x64",
    "Macintosh; Intel Mac OS X 10_10_5",
    "iPad; CPU OS 8_4_1 like Mac OS X",
    "Linux; U; Android 4.4.3; en-us"
]
ENGINES: list[str] = [
    "AppleWebKit/537.36 (KHTML, like Gecko)",
    "AppleWebKit/600.8.9 (KHTML, like Gecko)",
    "Gecko/20100101",
    "Trident/7.0; rv:11.0"
]
BROWSERS: list[str] = [
    "Chrome/45.0.2454.85 Safari/537.36",
    "Chromium/37.0.2062.94 Chrome/37.0.2062.94 Safari/537.36",
    "Firefox/40.0",
    "Version/8.0.8 Safari/600.8.9",
    "Mobile/12H321 Safari/600.1.4"
]


def generate_user_agent() -> str:
    return f"Mozilla/5.0 ({random.choice(PLATFORMS)}) {random.choice(ENGINES)} {random.choice(BROWSERS)}"



from __future__ import annotations

from typing import Self

from playwright.async_api import Browser, BrowserContext, async_playwright, Playwright

from .utils import generate_user_agent


class Generator:
    def __init__(self) -> None:
        super().__init__()

        self._pw: Playwright | None = None
        self._browser: Browser | None = None
        self._context: BrowserContext | None = None

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        await self.close()

    async def _start(self) -> None:
        if not self._pw:
            self._pw = await async_playwright().start()
        if not self._browser:
            self._browser = await self._pw.chromium.launch(headless=True)
        if not self._context:
            self._context = await self._browser.new_context(
                user_agent=generate_user_agent()
            )

    async def close(self) -> None:
        """Close the generator and release resources."""
        if self._context:
            await self._context.close()
        if self._browser:
            await self._browser.close()
        if self._pw:
            await self._pw.stop()


from __future__ import annotations

import asyncio
import json
import random
from typing import AsyncGenerator

from . import errors
from .generator import Generator


class TextGenerator(Generator):
    """AI text generator"""

    BASE_URL = "https://text-generation.perchance.org/api"

    def __init__(self) -> None:
        super().__init__()

        self._lock: asyncio.Lock = asyncio.Lock()

    def is_running(self) -> bool:
        return self._lock.locked()

    async def stream(
        self,
        prompt: str,
        *,
        start_with: str | None = None,
        stop_sequences: list[str] | None = None,
        timeout: float | None = 5.0,
    ) -> AsyncGenerator[str, None]:
        """Stream generated text.

        Parameters
        ----------
        prompt: str
            The prompt to generate text from.
        start_with: str | None
            Text to start the generation with.
        stop_sequences: list[str] | None
            List of sequences to stop the generation at.
        timeout: float | None
            Waiting timeout in seconds.
        """
        async with self._lock:
            await self._start()

            async with await self._context.new_page() as page:
                await page.goto(
                    f"{self.BASE_URL}/verifyUser"
                    f"?thread=0"
                    f"&__cacheBust={random.random()}"
                )

                content = await page.content()
                key_entry = content.find('"userKey":"')
                start_index = key_entry + len('"userKey":"')
                end_index = content.find('"', start_index)

                if key_entry == -1 or end_index == -1:
                    if "too_many_requests" in content:
                        raise errors.RateLimitError("Rate limit exceeded")
                    else:
                        raise errors.AuthenticationError("Failed to retrieve user key")

                key = content[start_index:end_index]

                url = (
                    f"{self.BASE_URL}/generate"
                    f"?userKey={key}"
                    f"&requestId=aiTextCompletion{random.randint(0, 2**30)}"
                    f"&__cacheBust={random.random()}"
                )
                body = {
                    "generatorName": "ai-text-generator",
                    "instruction": prompt,
                    "instructionTokenCount": 1,
                    "startWith": start_with or "",
                    "startWithTokenCount": 1,
                    "stopSequences": stop_sequences or [],
                }

                queue: asyncio.Queue[str] = asyncio.Queue()

                await page.expose_function("onChunk", queue.put)

                fetch_task = asyncio.create_task(page.evaluate("""
                    async ({ url, body }) => {
                        const controller = new AbortController();
                        window.abortFetch = () => controller.abort();

                        const response = await fetch(url, {
                            method: 'POST',
                            headers: { 'Content-Type': 'application/json' },
                            body: JSON.stringify(body),
                            signal: controller.signal
                        });

                        const reader = response.body.getReader();
                        const decoder = new TextDecoder();

                        while (true) {
                            const { value, done } = await reader.read();
                            if (done) break;
                            const chunk = decoder.decode(value, { stream: true });
                            await window.onChunk(chunk);
                        }
                    }
                """, {"url": url, "body": body}))

                while True:
                    try:
                        chunk = await asyncio.wait_for(queue.get(), timeout=timeout)
                        for line in chunk.splitlines():
                            if line.startswith("t:"):
                                yield json.loads(line[2:])
                            elif line.startswith("data:"):
                                return
                    except asyncio.TimeoutError:
                        await page.evaluate("window.abortFetch()")
                        fetch_task.cancel()
                        raise errors.ConnectionError()

    async def text(
        self,
        prompt: str,
        *,
        start_with: str | None = None,
        stop_sequences: list[str] | None = None,
        timeout: float | None = 5.0,
    ) -> str:
        """Generate text.

        Parameters
        ----------
        prompt: str
            The prompt to generate text from.
        start_with: str | None
            Text to start the generation with.
        stop_sequences: list[str] | None
            List of sequences to stop the generation at.
        timeout: float | None
            Waiting timeout in seconds.
        """
        result = []
        async for chunk in self.stream(
            prompt,
            start_with=start_with,
            stop_sequences=stop_sequences,
            timeout=timeout,
        ):
            result.append(chunk)

        return "".join(result)



from __future__ import annotations

import base64
import io
import random
from typing import Literal

import aiofiles

from . import errors
from .generator import Generator


class ImageResult:
    """Image generation result."""

    def __init__(
        self, 
        *, 
        generator: ImageGenerator,
        image_id: str,
        file_extension: str,
        seed: int,
        prompt: str,
        width: int,
        height: int,
        guidance_scale: float,
        negative_prompt: str | None,
        maybe_nsfw: bool
    ) -> None:
        self._generator: ImageGenerator = generator

        self.image_id: str = image_id
        """Image ID."""
        self.file_extension: str = file_extension
        """File extension."""
        self.seed: int = seed
        """Generation seed."""
        self.prompt: str = prompt
        """Image prompt."""
        self.width: int = width
        """Image width."""
        self.height: int = height
        """Image height."""
        self.guidance_scale: float = guidance_scale
        """Guidance scale."""
        self.negative_prompt: str | None = negative_prompt
        """Negative prompt."""
        self.maybe_nsfw: bool = maybe_nsfw
        """Whether the image may be NSFW."""
    
    def __str__(self) -> str:
        return f"{self.image_id}.{self.file_extension}"

    @property
    def size(self) -> tuple[int, int]:
        """Image size as (width, height)."""
        return self.width, self.height

    async def download(self) -> io.BytesIO:
        """Download the image."""
        url = (
            f"{self._generator.BASE_URL}/downloadTemporaryImage"
            f"?imageId={self.image_id}"
        )

        async with await self._generator._context.new_page() as page:
            await page.goto(
                f"{self._generator.BASE_URL}/verifyUser"
                f"?thread=0"
                f"&__cacheBust={random.random()}"
            )
            
            response_data = await page.evaluate("""
                async (url) => {
                    const response = await fetch(url);
                    if (!response.ok) {
                        return { ok: false, status: response.status };
                    }

                    const blob = await response.blob();

                    const base64 = await new Promise(resolve => {
                        const reader = new FileReader();
                        reader.onloadend = () => {
                            resolve(reader.result.split(",")[1]);
                        };
                        reader.readAsDataURL(blob);
                    });

                    return { ok: true, data: base64 };
                }
                """, url)
                            
            if not response_data['ok']:
                raise errors.ConnectionError(
                    f"Failed to download image: {response_data['status']}"
                )
            
            data = base64.b64decode(response_data['data'])
            return io.BytesIO(data)
 
    async def save(self, filename: str | None = None) -> None:
        """Download and save the image.

        Parameters
        ----------
        filename: str | None
            Name of the output file.
        """
        file = filename or f"{self.image_id}.{self.file_ext}" 

        async with aiofiles.open(file, 'wb') as f:
            img = await self.download()
            await f.write(img.read())
            

class ImageGenerator(Generator):
    """AI image generator"""

    BASE_URL = "https://image-generation.perchance.org/api"

    def __init__(self) -> None:
        super().__init__()

    async def image(
        self,
        prompt: str,
        *,
        negative_prompt: str | None = None,
        seed: int = -1,
        shape: Literal['portrait', 'square', 'landscape'] = 'square',
        guidance_scale: float = 7.0
    ) -> ImageResult:
        """
        Generate image.

        Parameters
        ----------
        prompt: str
            Image description.
        negative_prompt: str | None
            Things you do NOT want to see in the image.
        seed: int
            Generation seed.
        shape: str
            Image shape. Can be either `portrait`, `square` or `landscape`.
        guidance_scale: float
            Accuracy of the prompt in range `1-30`. 
        """
        if shape == 'portrait':
            resolution = '512x768'
        elif shape == 'square':
            resolution = '768x768'
        elif shape == 'landscape':
            resolution = '768x512'
        else:
            raise ValueError(f"Invalid shape: {shape}")
      
        await self._start()

        async with await self._context.new_page() as page:
            await page.goto(
                f"{self.BASE_URL}/verifyUser"
                f"?thread=0"
                f"&__cacheBust={random.random()}"
            )

            content = await page.content()

            key_entry = content.find('"userKey":"')
            start_index = key_entry + len('"userKey":"')
            end_index = content.find('"', start_index)

            if key_entry == -1 or end_index == -1:
                if "too_many_requests" in content:
                    raise errors.RateLimitError("Rate limit exceeded")
                else:
                    raise errors.AuthenticationError("Failed to retrieve user key")

            key = content[start_index:end_index]

            url = (
                f"{self.BASE_URL}/generate"
                f"?userKey={key}"
                f"&requestId=aiImageCompletion{random.randint(0, 2**30)}"
                f"&__cacheBust={random.random()}"
            )
            body = {
                "generatorName": "ai-image-generator",
                "channel": "ai-text-to-image-generator",
                "subChannel": "public",
                "prompt": prompt,
                "negativePrompt": negative_prompt or "",
                "seed": seed,
                "resolution": resolution,
                "guidanceScale": guidance_scale
            }
      
            response = await page.evaluate("""
                async ({ url, body }) => {
                    const controller = new AbortController();
                    window.abortFetch = () => controller.abort();

                    const response = await fetch(url, {
                        method: 'POST',
                        headers: { 'Content-Type': 'application/json' },
                        body: JSON.stringify(body),
                        signal: controller.signal
                    });
                                
                    return await response.json();
                }
            """, {"url": url, "body": body})

            return ImageResult(
                generator=self,
                image_id=response['imageId'],
                file_extension=response['fileExtension'],
                seed=response['seed'],
                prompt=response['prompt'],
                width=response['width'],
                height=response['height'],
                guidance_scale=response['guidanceScale'],
                negative_prompt=response['negativePrompt'],
                maybe_nsfw=response['maybeNsfw']
            )
                        